import pytest
from yarl import URL

from repoproviders.resolvers.base import NotFound
from repoproviders.resolvers.doi import DataverseDataset, DataverseResolver


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com/something", None),
        # Dataverse URLs should wholly under the subpath URL
        ("https://dvn.library.ubc.ca/not-dvn/something-else", None),
        # A dataset citation returns the dataset correctly
        (
            "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/TJCLKP",
            DataverseDataset(
                URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/TJCLKP"
            ),
        ),
        (
            "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/TJCLKP",
            DataverseDataset(
                URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/TJCLKP"
            ),
        ),
        # Asking for specific files should give us the whole dataset they are a part of
        (
            "https://dataverse.harvard.edu/api/access/datafile/3323458",
            DataverseDataset(
                URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/3MJ7IR"
            ),
        ),
        (
            "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ",
            DataverseDataset(
                URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/6ZXAGT"
            ),
        ),
        (
            "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ",
            DataverseDataset(
                URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/6ZXAGT"
            ),
        ),
        # Asking for datasets that don't exist should return NotFound
        (
            "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/not-found",
            NotFound(),
        ),
        (
            "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/not-found",
            NotFound(),
        ),
        ("https://dataverse.harvard.edu/api/access/datafile/0", NotFound()),
        ("https://dataverse.harvard.edu/blaaaah", NotFound()),
        (
            "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/not-found",
            NotFound(),
        ),
    ),
)
async def test_dataverse(url, expected):
    dv = DataverseResolver()
    assert await dv.resolve(URL(url)) == expected



from dataclasses import dataclass
from typing import Any, Protocol


@dataclass(frozen=True)
class NotFound:
    """
    Resolver recognizes this question, but while resolving determined it does not exist
    """

    immutable = True


class Repo(Protocol):
    """
    Represents a Repository
    """

    # Set to true if the repo is identified to be immutable. This usually means one of:
    #
    # 1. The identifier used to refer to the repo is guaranteed to not change (like with zenodo record ids)
    # 2. A version identifier is included as part of the definition (like with Figshare)
    # 3. A content identifiable hash is included as part of the definition (like with ImmutableGit)
    immutable: bool


class SupportsResolve(Protocol):
    async def resolve(self, question: Any) -> Repo | None:
        pass



import inspect
import types
import typing
from typing import Any

from yarl import URL

from .base import Repo, SupportsResolve
from .doi import (
    DataverseResolver,
    DoiResolver,
    FigshareResolver,
    ImmutableFigshareResolver,
    ZenodoResolver,
)
from .git import GitHubResolver, GitUrlResolver, ImmutableGitResolver

ALL_RESOLVERS: list[SupportsResolve] = [
    GitHubResolver(),
    GitUrlResolver(),
    DoiResolver(),
    ZenodoResolver(),
    FigshareResolver(),
    ImmutableFigshareResolver(),
    DataverseResolver(),
    ImmutableGitResolver(),
]

RESOLVER_BY_TYPE: dict[type, list[SupportsResolve]] = {}
for R in ALL_RESOLVERS:
    annotations = inspect.get_annotations(R.resolve)
    supported_types = annotations["question"]
    if isinstance(supported_types, type):
        # Only supports a single type
        RESOLVER_BY_TYPE.setdefault(supported_types, []).append(R)
    elif isinstance(supported_types, types.UnionType):
        for t in typing.get_args(supported_types):
            RESOLVER_BY_TYPE.setdefault(t, []).append(R)


async def resolve(question: str | Any, recursive: bool) -> list[Repo] | None:
    if isinstance(question, str):
        question = URL(question)

    answers = []
    resp = None

    while True:
        # Get a list of applicable resolvers we can try
        applicable_resolvers = RESOLVER_BY_TYPE.get(type(question), [])

        if not applicable_resolvers:
            # No applicable resolvers found for this question type, we are done
            break

        for r in applicable_resolvers:
            resp = await r.resolve(question)
            if resp is not None:
                # We found an answer!
                answers.append(resp)
                # Break after we find an answer in each round
                break

        if recursive:
            # If we want a recursive answer, we have to continue iterating
            if resp is not None:
                # We *did* find an answer this round, so we should continue and see if we find more
                question = resp
                resp = None
            else:
                # We did *not* find an answer this round, so we are done
                break
        else:
            # We are not recursive, so we are done after 1 round regardless
            break

    return answers



import inspect
import types
import typing
from pathlib import Path
from typing import Any

from .base import SupportsFetch
from .dataverse import DataverseFetcher
from .git import ImmutableGitFetcher

ALL_FETCHERS: list[SupportsFetch] = [ImmutableGitFetcher(), DataverseFetcher()]

FETCHER_BY_TYPE: dict[type, SupportsFetch] = {}
for R in ALL_FETCHERS:
    annotations = inspect.get_annotations(R.fetch)
    supported_types = annotations["repo"]
    if isinstance(supported_types, type):
        # Only supports a single type
        FETCHER_BY_TYPE[supported_types] = R
    elif isinstance(supported_types, types.UnionType):
        for t in typing.get_args(supported_types):
            FETCHER_BY_TYPE[t] = R


async def fetch(question: Any, output_dir: Path):
    fetcher = FETCHER_BY_TYPE[type(question)]

    await fetcher.fetch(question, output_dir)



import asyncio
import re
from dataclasses import dataclass

from yarl import URL

from .base import NotFound


@dataclass(frozen=True)
class Git:
    repo: str
    ref: str

    immutable = False


@dataclass(frozen=True)
class ImmutableGit:
    """
    Same as Git, but marked to be fully resolved. This implies:

    1. The repository exists, and can be contacted
    2. If ref was a branch or tag, it has been resolved into an immutable commit sha
    3. If ref *looks* like a sha, we assume it exists (without testing it)
    """

    repo: str
    ref: str

    immutable = True


class GitHubResolver:
    async def resolve(self, question: URL) -> Git | None:
        # git+<scheme> urls are handled by a different resolver
        if question.scheme not in ("http", "https") or (
            question.host != "github.com" and question.host != "www.github.com"
        ):
            # TODO: Allow configuring for GitHub enterprise
            return None

        # Split the URL into parts, discarding empty parts to account for leading and trailing slashes
        parts = [p for p in question.path.split("/") if p.strip() != ""]
        if len(parts) == 2:
            # Handle <user|org>/<repo>
            # Reconstruct the URL so we normalize any
            return Git(
                repo=str(question.with_path(f"{parts[0]}/{parts[1]}")), ref="HEAD"
            )
        elif len(parts) >= 4 and parts[2] in ("tree", "blob"):
            # Handle <user|org>/<repo>/<tree|blob>/<ref>(/<possible-path>)
            # Note: We ignore any paths specified here, as we only care about the repo
            return Git(
                repo=str(question.with_path(f"{parts[0]}/{parts[1]}")), ref=parts[3]
            )
        else:
            # This is not actually a valid GitHub URL we support
            return None


class ImmutableGitResolver:
    async def resolve(self, question: Git) -> ImmutableGit | NotFound | None:
        command = ["git", "ls-remote", "--", question.repo, question.ref]
        proc = await asyncio.create_subprocess_exec(
            *command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )
        stdout, stderr = [s.decode().strip() for s in await proc.communicate()]
        retcode = await proc.wait()
        if retcode:
            # `git` may follow redirects here, so the repo we pass may not always be the repo we
            # get back. So we loosely check for a 'not found' message.
            if re.search(r"fatal: repository '(.+)' not found", stderr, re.MULTILINE):
                return NotFound()

            # If it's another error, let's raise it directly
            raise RuntimeError(
                f"Unable to run git ls-remote to resolve {question}: {stderr}"
            )

        if stdout == "":
            # The remote repo exists, and we can talk to it. But no *ref* with the given name
            # exists. We check if this looks like a commit sha, and if it is, assume it exists.
            # FIXME: Decide if we just do this check *before* trying ls-remotes? Faster, but it means
            # we can't guarantee that the repo itself exists
            if re.match(r"[0-9a-f]{40}", question.ref):
                resolved_ref = question.ref
            else:
                return NotFound()
        else:
            resolved_ref = stdout.split("\t", 1)[0]

        return ImmutableGit(question.repo, resolved_ref)


class GitUrlResolver:
    """
    Resolves raw git URLs

    URL structure is inspired by what `pip` supports: https://pip.pypa.io/en/stable/topics/vcs-support/#git
    """

    async def resolve(self, question: URL) -> Git | None:
        # List of supported protocols is from https://pip.pypa.io/en/stable/topics/vcs-support/#git
        if question.scheme not in (
            "git+https",
            "git+ssh",
            "git",
            "git+file",
            "git+http",
            "git+git",
        ):
            return None

        repo = question.with_scheme(question.scheme.replace("git+", ""))

        if "@" in question.path:
            parts = question.path.split("@", 1)
            ref = parts[1]
            repo = repo.with_path(parts[0])
        else:
            ref = "HEAD"

        return Git(str(repo), ref)



import pytest
from yarl import URL

from repoproviders.resolvers.base import NotFound
from repoproviders.resolvers.doi import ZenodoDataset, ZenodoResolver


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com/something", None),
        # A non-dataset URL
        ("https://data.caltech.edu/communities", None),
        # Simple /record and /records
        (
            "https://zenodo.org/record/3232985",
            ZenodoDataset(URL("https://zenodo.org/"), "3232985"),
        ),
        (
            "https://zenodo.org/records/3232985",
            ZenodoDataset(URL("https://zenodo.org/"), "3232985"),
        ),
        # Note we normalize output to have the HTTPS URL, even if we're passed in the HTTP URL
        (
            "http://zenodo.org/record/3232985",
            ZenodoDataset(URL("https://zenodo.org/"), "3232985"),
        ),
        (
            "https://zenodo.org/records/3232985",
            ZenodoDataset(URL("https://zenodo.org/"), "3232985"),
        ),
        # A non-zenodo URL
        (
            "https://data.caltech.edu/records/996aw-mf266",
            ZenodoDataset(URL("https://data.caltech.edu/"), "996aw-mf266"),
        ),
        # A doi reference
        (
            "https://zenodo.org/doi/10.5281/zenodo.805993",
            ZenodoDataset(URL("https://zenodo.org/"), recordId="14007206"),
        ),
        # A doi reference to a bad doi
        ("https://zenodo.org/doi/10.5281/zdo.805993", NotFound()),
    ),
)
async def test_zenodo(url, expected):
    zr = ZenodoResolver()
    assert await zr.resolve(URL(url)) == expected



import pytest
from yarl import URL

from repoproviders.resolvers import resolve
from repoproviders.resolvers.base import NotFound
from repoproviders.resolvers.doi import (
    DataverseDataset,
    Doi,
    FigshareDataset,
    FigshareInstallation,
    ImmutableFigshareDataset,
    ZenodoDataset,
)
from repoproviders.resolvers.git import Git, ImmutableGit


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com/something", []),
        # GitHub URLs that are not repos
        ("https://github.com/pyOpenSci", []),
        (
            "https://github.com/yuvipanda/repoproviders/actions/runs/12552733471/job/34999118812",
            [],
        ),
        ("https://github.com/yuvipanda/repoproviders/settings", []),
        ("https://github.com/jupyter/docker-stacks/pull/2194", []),
        # Simple github repo URL
        (
            "https://github.com/pyOpenSci/pyos-package-template",
            [Git("https://github.com/pyOpenSci/pyos-package-template", "HEAD")],
        ),
        # Trailing slash normalized?
        (
            "https://github.com/pyOpenSci/pyos-package-template/",
            [Git("https://github.com/pyOpenSci/pyos-package-template", "HEAD")],
        ),
        # blobs and tree
        (
            "https://github.com/pyOpenSci/pyos-package-template/tree/main/includes/licenses",
            [Git("https://github.com/pyOpenSci/pyos-package-template", "main")],
        ),
        (
            "https://github.com/pyOpenSci/pyos-package-template/tree/original-cookie/docs",
            [
                Git(
                    "https://github.com/pyOpenSci/pyos-package-template",
                    "original-cookie",
                )
            ],
        ),
        (
            "https://github.com/pyOpenSci/pyos-package-template/blob/b912433bfae541972c83529359f4181ef0fe9b67/README.md",
            [
                Git(
                    "https://github.com/pyOpenSci/pyos-package-template",
                    ref="b912433bfae541972c83529359f4181ef0fe9b67",
                )
            ],
        ),
        # Random URL, not a git repo
        (Git("https://example.com/something", "HEAD"), [NotFound()]),
        # Resolve a tag
        (
            Git("https://github.com/jupyterhub/zero-to-jupyterhub-k8s", "0.8.0"),
            [
                ImmutableGit(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "ada2170a2181ae1760d85eab74e5264d0c6bb67f",
                )
            ],
        ),
        # Resolve a commit we know exists, although this isn't verified
        (
            Git(
                "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
            ),
            [
                ImmutableGit(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                )
            ],
        ),
        (
            Git(
                "https://github.com/jupyterhub/zero-to-jupyterhub-k8s", "does-not-exist"
            ),
            [NotFound()],
        ),
    ),
)
async def test_resolve(url, expected):
    assert await resolve(url, False) == expected


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com/something", []),
        # doi schema'd URI
        (
            "doi:10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Doi(
                    "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                )
            ],
        ),
        # handle schema'd URI
        (
            "hdl:11529/10016",
            [Doi("https://data.cimmyt.org/dataset.xhtml?persistentId=hdl:11529/10016")],
        ),
        # For convenience, we do accept DOIs without a scheme
        (
            "10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Doi(
                    "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                )
            ],
        ),
        # But not handles without a scheme
        ("11529/10016", []),
        # Three DOI resolution URLs
        (
            "https://doi.org/10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Doi(
                    "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                )
            ],
        ),
        (
            "https://www.doi.org/10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Doi(
                    "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                )
            ],
        ),
        (
            "https://hdl.handle.net/10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Doi(
                    "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                )
            ],
        ),
    ),
)
async def test_norecurse(url, expected):
    assert await resolve(URL(url), False) == expected


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # doi schema'd URI
        (
            "doi:10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Doi(
                    "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                ),
                DataverseDataset(
                    URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/6ZXAGT"
                ),
            ],
        ),
        # handle schema'd URI
        (
            "hdl:11529/10016",
            [
                Doi(
                    "https://data.cimmyt.org/dataset.xhtml?persistentId=hdl:11529/10016"
                ),
                DataverseDataset(URL("http://data.cimmyt.org/"), "hdl:11529/10016"),
            ],
        ),
        # For convenience, we do accept DOIs without a scheme
        (
            "10.7910/DVN/6ZXAGT",
            [
                Doi(
                    "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/6ZXAGT"
                ),
                DataverseDataset(
                    URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/6ZXAGT"
                ),
            ],
        ),
        # Something that's only a DOI, and won't resolve further
        (
            "10.1126/science.aar3646",
            [Doi("https://www.science.org/doi/10.1126/science.aar3646")],
        ),
        # GitHub URLs that recurse into ImmutableGit
        (
            "https://github.com/jupyterhub/zero-to-jupyterhub-k8s/tree/f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
            [
                Git(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                ),
                ImmutableGit(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                ),
            ],
        ),
        (
            "https://github.com/jupyterhub/zero-to-jupyterhub-k8s/tree/0.8.0",
            [
                Git(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "0.8.0",
                ),
                ImmutableGit(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "ada2170a2181ae1760d85eab74e5264d0c6bb67f",
                ),
            ],
        ),
        # Git URLs that recurse into ImmutableGit
        (
            "git+https://github.com/jupyterhub/zero-to-jupyterhub-k8s@f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
            [
                Git(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                ),
                ImmutableGit(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                ),
            ],
        ),
        (
            "git+https://github.com/jupyterhub/zero-to-jupyterhub-k8s@0.8.0",
            [
                Git(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "0.8.0",
                ),
                ImmutableGit(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "ada2170a2181ae1760d85eab74e5264d0c6bb67f",
                ),
            ],
        ),
        (
            "10.5281/zenodo.3232985",
            [
                Doi("https://zenodo.org/record/3232985"),
                ZenodoDataset(URL("https://zenodo.org/"), "3232985"),
            ],
        ),
        (
            "https://doi.org/10.6084/m9.figshare.9782777.v3",
            [
                Doi(
                    "https://figshare.com/articles/Binder-ready_openSenseMap_Analysis/9782777/3"
                ),
                FigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    3,
                ),
                ImmutableFigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    3,
                ),
            ],
        ),
        (
            "https://figshare.com/articles/Binder-ready_openSenseMap_Analysis/9782777",
            [
                FigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    None,
                ),
                ImmutableFigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    3,
                ),
            ],
        ),
        # A zenodo doi that uses the /doi redirect
        (
            "https://doi.org/10.5281/zenodo.805993",
            [
                Doi(url="https://zenodo.org/doi/10.5281/zenodo.805993"),
                ZenodoDataset(URL("https://zenodo.org/"), "14007206"),
            ],
        ),
    ),
)
async def test_recurse(url, expected):
    assert await resolve(URL(url), True) == expected



import pytest
from yarl import URL

from repoproviders.resolvers.git import Git, GitHubResolver


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com/something", None),
        # GitHub URLs that are not repos
        ("https://github.com/pyOpenSci", None),
        (
            "https://github.com/yuvipanda/repoproviders/actions/runs/12552733471/job/34999118812",
            None,
        ),
        ("https://github.com/yuvipanda/repoproviders/settings", None),
        ("https://github.com/jupyter/docker-stacks/pull/2194", None),
        # Simple github repo URL
        (
            "https://github.com/pyOpenSci/pyos-package-template",
            Git("https://github.com/pyOpenSci/pyos-package-template", "HEAD"),
        ),
        # Trailing slash normalized?
        (
            "https://github.com/pyOpenSci/pyos-package-template/",
            Git("https://github.com/pyOpenSci/pyos-package-template", "HEAD"),
        ),
        # blobs and tree
        (
            "https://github.com/pyOpenSci/pyos-package-template/tree/main/includes/licenses",
            Git("https://github.com/pyOpenSci/pyos-package-template", "main"),
        ),
        (
            "https://github.com/pyOpenSci/pyos-package-template/tree/original-cookie/docs",
            Git(
                "https://github.com/pyOpenSci/pyos-package-template", "original-cookie"
            ),
        ),
        (
            "https://github.com/pyOpenSci/pyos-package-template/blob/b912433bfae541972c83529359f4181ef0fe9b67/README.md",
            Git(
                "https://github.com/pyOpenSci/pyos-package-template",
                ref="b912433bfae541972c83529359f4181ef0fe9b67",
            ),
        ),
        (
            "https://github.com/yuvipanda/does-not-exist-e43",
            Git(repo="https://github.com/yuvipanda/does-not-exist-e43", ref="HEAD"),
        ),
    ),
)
async def test_github(url, expected):
    gh = GitHubResolver()
    assert await gh.resolve(URL(url)) == expected
